local folder, core = ...

if not core.LibNameplate then
	return
end

local L = core.L or LibStub("AceLocale-3.0"):GetLocale(folder, true)
local _G = _G
local pairs = pairs
local GetTime = GetTime
local CreateFrame = CreateFrame
local table_remove = table.remove
local table_sort = table.sort
local type = type
local table_getn = table.getn
local Debug = core.Debug
local table_getn = table.getn
local DebuffTypeColor = DebuffTypeColor
local select = select
local string_gsub = string.gsub

local P = {}
local LibNameplate = core.LibNameplate
local nametoGUIDs = core.nametoGUIDs
local buffBars = core.buffBars
local buffFrames = core.buffFrames
local guidBuffs = core.guidBuffs


local prev_OnEnable = core.OnEnable
function core:OnEnable()
	prev_OnEnable(self)
	
--~ 	Debug("OnEnable", "frames", self.db.profile)
	P = self.db.profile --this can change on profile change.
end

----------------------------------------------------------
local function UpdateBuffSize(frame, size)				--
-- Update a spell frame's texture size. 				--
-- This takes into adjusts height for cooldown size.	--
----------------------------------------------------------
	frame.texture:SetWidth(size)
	frame.texture:SetHeight(size)
	
	frame.border:SetHeight(size)
	frame.border:SetWidth(size)
	
	frame:SetWidth(size)
	
	if P.showCooldown == true then
		frame:SetHeight(size + frame.cd:GetStringHeight())
	else
		frame:SetHeight(size)
	end
end

------------------------------------------------------
local function UpdateBuffCDSize(buffFrame, size)	--
-- Set cooldown text size.							--
------------------------------------------------------
	buffFrame.cd:SetFont("Fonts\\FRIZQT__.TTF", size, "NORMAL")
	buffFrame.cdbg:SetHeight(buffFrame.cd:GetStringHeight())
--~ 	iconOnShow(buffFrame) --lazy way to hide/show cooldown.
end

--------------------------------------------------
local function SetStackSize(buffFrame, size)	--
-- Set the stack text size.						--
--------------------------------------------------
	buffFrame.stack:SetFont("Fonts\\FRIZQT__.TTF", size, "OUTLINE")
end

------------------------------------------
local function iconOnShow(self)			--
-- Called when spell frames are shown.	--
------------------------------------------
	self:SetAlpha(1)
	if P.showCooldown == true and self.expirationTime > 0 then
		self.cdbg:Show()
		self.cd:Show()
	else
		self.cdbg:Hide()
		self.cd:Hide()
	end
	
	local iconSize = P.iconSize
	local cooldownSize = P.cooldownSize
	local stackSize = P.stackSize
	local spellName = self.spellName or "X"

	if P.spellOpts[spellName] then
		iconSize = P.spellOpts[spellName].iconSize or iconSize
		cooldownSize = P.spellOpts[spellName].cooldownSize or cooldownSize
		stackSize = P.spellOpts[spellName].stackSize or stackSize
	end

	if self.playerCast and P.biggerSelfSpells == true then
		UpdateBuffSize(self, iconSize * 1.2)
	else
		UpdateBuffSize(self, iconSize)
	end
	
	UpdateBuffCDSize(self, cooldownSize)

	if self.stackCount and self.stackCount > 1 then
		self.stack:Show()
		SetStackSize(self, stackSize)
	else
		self.stack:Hide()
	end
	
	if self.isDebuff then
		local colour = DebuffTypeColor[self.debuffType or ""]
		if colour then
			self.border:SetVertexColor(colour.r, colour.g, colour.b)
			self.border:Show()
		end
	else
		self.border:Hide()
	end
end

------------------------------------------
local function iconOnHide(self)			--
-- Called when spell frames are shown.	--
------------------------------------------
	self.stack:Hide()
	UpdateBuffSize(self, P.iconSize)
end

----------------------------------------------
local function iconOnUpdate(self, elapsed)	--
-- Fires for spell frames.					--
----------------------------------------------
	self.lastUpdate = self.lastUpdate + elapsed
	if self.lastUpdate > 0.1 then --abit fast for cooldown flash.
		self.lastUpdate = 0

		if self.stackCount > 1 then
			self.stack:SetText(self.stackCount)
		end
		
		if self.expirationTime > 0 then
			local timeLeft = core:Round(self.expirationTime - GetTime())
			if P.showCooldown == true then
				self.cd:SetText(core:SecondsToString(timeLeft, 1))
				self.cd:SetTextColor(core:RedToGreen(timeLeft, self.duration))
				self.cdbg:SetWidth(self.cd:GetStringWidth())
			end
			
			if (timeLeft / self.duration ) < P.blinkTimeleft and timeLeft < 60 then --buff only has 20% timeleft and is less then 60 seconds.
				local f = GetTime() % 1
				if f > 0.5 then 
					f = 1 - f
				end
				self:SetAlpha(2 * f)
			end
			
			if GetTime() > self.expirationTime then
				self:Hide()
				
				local GUID = LibNameplate:GetGUID(self.realPlate)
				if GUID then
					core:AddBuffsToPlate(self.realPlate, GUID)
				else
					local plateName =  LibNameplate:GetName(self.realPlate)
					if plateName and nametoGUIDs[plateName] then
						core:AddBuffsToPlate(self.realPlate, nametoGUIDs[plateName])
					end
				end
			end
--~ 			else
--~ 				self.cd:Hide()
--~ 				self.cdbg:Hide()
		end
		
	end
end





local function SetBarSize(barFrame, width, height)
	barFrame:SetWidth(width)
	barFrame:SetHeight(height)
end



local function CreateBuffFrame(parentFrame, realPlate)
	local f = CreateFrame("frame", nil, parentFrame) --
	f.realPlate = realPlate
	f:SetFrameStrata("LOW")
	
	f.texture = f:CreateTexture(nil,"LOW")
	f.texture:SetPoint("TOP", f)


	
	local cd = f:CreateFontString(nil, "ARTWORK", "ChatFontNormal")
--~ 	cd:SetFont("Fonts\\FRIZQT__.TTF", P.cooldownSize, "NORMAL") --done later.
	
	
	cd:SetText("0")
	cd:SetPoint("TOP", f.texture, "BOTTOM")
	f.cd = cd

	--Make the text easier to see.
	f.cdbg = f:CreateTexture(nil,"LOW")
	f.cdbg:SetTexture(0,0,0,.75)
	f.cdbg:SetPoint("CENTER", cd)
	
	f.stack = f:CreateFontString(nil, "ARTWORK", "ChatFontNormal")

	f.stack:SetText("0")
	f.stack:SetPoint("BOTTOMRIGHT", f.texture, "BOTTOMRIGHT", -1, 3)
	f.stack:Hide()

	
	f.border = f:CreateTexture(nil,"OVERLAY")
	f.border:SetPoint("CENTER", f.texture)
	f.border:SetTexture("Interface\\Buttons\\UI-Debuff-Overlays")
	f.border:SetTexCoord(0.296875,0.5703125,0,0.515625)

	f.lastUpdate = 0
	f.expirationTime = 0
	f:SetScript("OnShow", iconOnShow)
	f:SetScript("OnHide", iconOnHide)
	
	
	f:SetScript("OnUpdate", iconOnUpdate) 
	f.stackCount = 0

	return f
end

------------------------------------------
function core:UpdateBarsBackground()	--
-- Show/Hide bar background texture.	--
------------------------------------------
	for plate in pairs(buffBars) do 
		for b in pairs(buffBars[plate]) do 
		
			if P.showBarBackground == true then
				buffBars[plate][b].barBG:Show()
			else
				buffBars[plate][b].barBG:Hide()
			end
		
		end
	end
end
--~ local barSizes = {}

----------------------------------------------------------
local function CreateBarFrame(parentFrame, realPlate)	--
-- Create and return a bar frame.						--
----------------------------------------------------------
	local f = CreateFrame("frame", nil, parentFrame) 
	f.realPlate = realPlate
	
	f:SetFrameStrata("BACKGROUND")
	f:SetFrameLevel(parentFrame:GetFrameLevel())
	
	f:SetWidth(1)
	f:SetHeight(1)
	
	--Make the text easier to see.
	f.barBG = f:CreateTexture(nil,"LOW")
	f.barBG:SetAllPoints(true)

	f.barBG:SetTexture(1,1,1,0.3) --uses for testing
	if P.showBarBackground == true then
		f.barBG:Show()
	else
		f.barBG:Hide()
	end
	
	f:Show()
	return f
end

----------------------------------------------------------------------
local function BuildPlateBars(plate, visibleFrame)					--
-- Build all our bar frames for a plate. 							--
-- We anchor these to the plate and our spell frames to the bar.	--
----------------------------------------------------------------------
	buffBars[plate] = buffBars[plate] or {}
	
	
	if not buffBars[plate][1] then
		buffBars[plate][1] = CreateBarFrame(visibleFrame, plate)
	end
	buffBars[plate][1]:ClearAllPoints()
	buffBars[plate][1]:SetPoint(P.barAnchorPoint, visibleFrame, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY) -- + GetCooldownHeight(buffFrames[plate][1])
	
	local barPoint		= P.barAnchorPoint
	local parentPoint	= P.plateAnchorPoint
	if P.barGrowth == 1 then --up
		barPoint = string_gsub(barPoint, "TOP", "BOTTOM");
		parentPoint = string_gsub(parentPoint, "BOTTOM", "TOP");
	else
		barPoint = string_gsub(barPoint, "BOTTOM,", "TOP");
		parentPoint = string_gsub(parentPoint, "TOP", "BOTTOM");
	end

	if P.numBars > 1 then
		for r=2, P.numBars do
			if not buffBars[plate][r] then
				buffBars[plate][r] = CreateBarFrame(visibleFrame, plate)
			end
			buffBars[plate][r]:ClearAllPoints()

			buffBars[plate][r]:SetPoint(barPoint, buffBars[plate][r-1], parentPoint, 0, 0)
		end
	end
end

local function GetBarChildrenSize(n, ...)
	local frame
	local totalWidth = 1
	local totalHeight = 1
	if n > P.iconsPerBar then
		n = P.iconsPerBar
	end
    for i = 1, n do
		frame = select(i, ...)
		if P.shrinkBar == true then
			if frame:IsShown() then
				totalWidth = totalWidth + frame:GetWidth()
				
				if frame:GetHeight() > totalHeight then
					totalHeight = frame:GetHeight()
				end
			end
		else
			totalWidth = totalWidth + frame:GetWidth()
			
			if frame:GetHeight() > totalHeight then
				totalHeight = frame:GetHeight()
			end
		end
    end
	return totalWidth, totalHeight
end

--------------------------------------------------------------------------------------
local function UpdateBarSize(barFrame)												--
-- Update a bar's size taking into account all the spell frame's height and width.	--
--------------------------------------------------------------------------------------
	if barFrame:GetNumChildren() == 0 then
--~ 		Debug("Our buff bar lost it's children?")
		return
	end

	local totalWidth, totalHeight = GetBarChildrenSize(barFrame:GetNumChildren(), barFrame:GetChildren())

	barFrame:SetWidth(totalWidth)
	barFrame:SetHeight(totalHeight)
end

local function UpdateAllBarSizes(plate)
	for r=P.numBars, 1, -1  do 
		UpdateBarSize(buffBars[plate][r])
	end
end

function core:UpdateAllPlateBarSizes()
	for plate in pairs(buffBars) do 
		UpdateAllBarSizes(plate)
	end
end

function core:RemoveOldSpells(GUID)
	for i=(P.numBars * P.iconsPerBar), 1, -1 do 
		if guidBuffs[GUID][i] then
			if guidBuffs[GUID][i].expirationTime > 0 and GetTime() > guidBuffs[GUID][i].expirationTime then
				table_remove(guidBuffs[GUID], i)
			end
		end
	end
end

----------------------------------------------
function core:AddBuffsToPlate(plate, GUID)	--
-- Show spells on a plate linked to a GUID.	--
----------------------------------------------

	if not buffFrames[plate] or not buffFrames[plate][P.numBars]  then
		self:BuildBuffFrame(plate)
	end
	
	local t, f
--~ 	Debug("AddBuffsToPlate 0", GUID)
	if guidBuffs[GUID] then
	
--~ 		Debug("AddBuffsToPlate 1", GUID, table.getn(guidBuffs[GUID]))
		
		self:RemoveOldSpells(GUID)
--~ 		Debug("AddBuffsToPlate 2", GUID, table.getn(guidBuffs[GUID]))
		
		table_sort(guidBuffs[GUID], function(a,b) 
			if(a and b) then 
				if a.playerCast ~= b.playerCast then
					return (a.playerCast or 0) > (b.playerCast or 0)
				elseif a.expirationTime == b.expirationTime then
					return a.name  < b.name
				else
					return (a.expirationTime or 0) < (b.expirationTime or 0)
				end
			end 
		end)

--~ 		Debug("AddBuffsToPlate 1", GUID)
		for i=1, P.numBars * P.iconsPerBar do 
--~ 			Debug("AddBuffsToPlate 2", GUID)
			if buffFrames[plate][i] then

				if guidBuffs[GUID][i] and (guidBuffs[GUID][i].expirationTime == 0 or guidBuffs[GUID][i].expirationTime > GetTime()) then
					buffFrames[plate][i].spellName		= guidBuffs[GUID][i].name or ""
					buffFrames[plate][i].expirationTime	= guidBuffs[GUID][i].expirationTime or 0
					buffFrames[plate][i].duration		= guidBuffs[GUID][i].duration or 1
					buffFrames[plate][i].stackCount = guidBuffs[GUID][i].stackCount or 0
					buffFrames[plate][i].isDebuff	= guidBuffs[GUID][i].isDebuff
					buffFrames[plate][i].debuffType = guidBuffs[GUID][i].debuffType
					buffFrames[plate][i].playerCast = guidBuffs[GUID][i].playerCast
					
					
					buffFrames[plate][i].texture:SetTexture("Interface\\Icons\\"..guidBuffs[GUID][i].icon)

					if buffFrames[plate][i]:IsShown() then
						iconOnShow(buffFrames[plate][i])
					else
						buffFrames[plate][i]:Show()--make sure OnShow fires.
					end
					iconOnUpdate(buffFrames[plate][i], 1) --do a onupdate.
				else
	--~ 				Debug("AddBuffsToPlate 4", i, GUID)
					buffFrames[plate][i]:Hide()
					buffFrames[plate][i].expirationTime = 0
				end
			end
		end

		UpdateAllBarSizes(plate)
	end


end

----------------------------------------------------------------------------------
function core:AddUnknownIcon(plate)												--
-- Display a question mark icon since we don't know the GUID of the plate/mob.	--
----------------------------------------------------------------------------------
	if not buffFrames[plate] then
		self:BuildBuffFrame(plate, nil, true)
	end

	local i = 1 --eaiser for me to copy/paste code elsewhere.
	buffFrames[plate][i].spellName		= false
	buffFrames[plate][i].expirationTime	= 0
	buffFrames[plate][i].duration		= 1
	buffFrames[plate][i].stackCount		= 0
	buffFrames[plate][i].isDebuff		= false
	buffFrames[plate][i].debuffType		= false
	buffFrames[plate][i].playerCast		= false
	
	
	buffFrames[plate][i].texture:SetTexture("Interface\\Icons\\Inv_misc_questionmark")
	
--~ 	Debug("AddUnknownIcon", "questionmark")

	if buffFrames[plate][i]:IsShown() then
		iconOnShow(buffFrames[plate][i])
	else
		buffFrames[plate][i]:Show()--make sure OnShow fires.
	end


	UpdateAllBarSizes(plate)

end

----------------------------------------------------------
function core:ResetAllPlateIcons()						--
-- This will reset all the anchors on the spell frames.	--
----------------------------------------------------------
	for plate in pairs(buffFrames) do 
		core:BuildBuffFrame(plate, true)
	end
end

------------------------------------------------------
function core:BuildBuffFrame(plate, reset, onlyOne)	--
-- Create our buff frames on a plate.				--
------------------------------------------------------

	local visibleFrame = LibNameplate:GetVisibleFrame(plate)
	if not buffBars[plate] then
		BuildPlateBars(plate, visibleFrame)
	end

	if not buffBars[plate][P.numBars] then--user increased the size.
		BuildPlateBars(plate, visibleFrame)
	end
	
	
	buffFrames[plate] = buffFrames[plate] or {}
	
	if reset then
		for i=1, table_getn(buffFrames[plate]) do 
			buffFrames[plate][i]:Hide() --makesure all frames stop OnUpdating.
		end
		
--~ 		for b in pairs(buffBars[plate]) do 
--~ 			buffBars[plate][b]:Hide()
--~ 		end
	end
	
	local total = 1 --total number of spell frames
	if not buffFrames[plate][total] then
		buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][1], plate)
	end
	buffFrames[plate][total]:SetParent(buffBars[plate][1])--Doing this seems to make UpdateBarSize() function properly. =/
	
	buffFrames[plate][total]:ClearAllPoints()
	buffFrames[plate][total]:SetPoint("TOPLEFT", buffBars[plate][1]) -- + GetCooldownHeight(buffFrames[plate][1])
	

	if onlyOne then
		--Save CPU cycles not creating every single frame. Lets wait till they're needed.
--~ 		UpdateBarSize(buffBars[plate][1])
		return
	end

	
	
	
	local prevFrame = buffFrames[plate][total]
	for i=2, P.iconsPerBar do 
		total = total + 1
		if not buffFrames[plate][total] then
			buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][1], plate)
		end
		buffFrames[plate][total]:SetParent(buffBars[plate][1])

		buffFrames[plate][total]:ClearAllPoints()
		buffFrames[plate][total]:SetPoint("TOPLEFT", prevFrame, "TOPRIGHT")

		prevFrame = buffFrames[plate][total]
	end
--~ 	UpdateBarSize(buffBars[plate][1])

	
	if P.numBars > 1 then
		for r=2, P.numBars do 
			for i=1, P.iconsPerBar do 
				total = total + 1

				if not buffFrames[plate][total] then
					buffFrames[plate][total] = CreateBuffFrame(buffBars[plate][r], plate)
				end
				buffFrames[plate][total]:SetParent(buffBars[plate][r])
				
				
				buffFrames[plate][total]:ClearAllPoints()
				if i == 1 then
					buffFrames[plate][total]:SetPoint("TOPLEFT", buffBars[plate][r])
				else
					buffFrames[plate][total]:SetPoint("TOPLEFT", prevFrame, "TOPRIGHT")
				end
				
				prevFrame = buffFrames[plate][total]
			end

--~ 			UpdateBarSize(buffBars[plate][r])
--~ 			buffBars[plate][r]:Show() --***

		end
	end

end

--------------------------------------------------------------
function core:ResetBarPoint(barFrame, plate, visibleFrame)	--
-- Reset a bar's anchor point.								--
--------------------------------------------------------------

	barFrame:ClearAllPoints()
	
--~ 			buffFrames[plate][i]:SetPoint(P.barAnchorPoint, plate, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY)
	local visibleFrame = visibleFrame or LibNameplate:GetVisibleFrame(plate)
	barFrame:SetParent(visibleFrame)
	barFrame:SetPoint(P.barAnchorPoint, visibleFrame, P.plateAnchorPoint, P.barOffsetX, P.barOffsetY) -- + GetCooldownHeight(barFrame)
end

--------------------------------------------------------------
function core:ResetIconSizes()								--
-- Reset all icon sizes. Called when user changes settings.	--
--------------------------------------------------------------

	local iconSize

	local frame
	for plate in pairs(self.buffFrames) do 
		for i=1, table_getn(self.buffFrames[plate]) do 
			frame = self.buffFrames[plate][i]
			
			
			if P.spellOpts[frame.name] then
				iconSize = P.spellOpts[frame.name].iconSize
			else
				iconSize = P.iconSize
			end
			
			--Update the spell texture size.
			frame.texture:SetWidth(iconSize)
			frame.texture:SetHeight(iconSize)

			--Update the frame as a whole, this takes into account the size of the cooldown size.
			frame:SetWidth(iconSize)
			
			if P.showCooldown == true then
				frame:SetHeight(iconSize + frame.cd:GetStringHeight())
			else
				frame:SetHeight(iconSize)
			end
		end
	end
end

----------------------------------------------------------------------
function core:ResetCooldownSize()									--
-- Reset cooldown text sizes. Called when user changes settings.	--
----------------------------------------------------------------------
	for plate in pairs(buffFrames) do 
		for i=1, table_getn(buffFrames[plate]) do 
--~ 			buffFrames[plate][i].cd:SetFontObject(core:GetCooldownFont())

			UpdateBuffCDSize(buffFrames[plate][i], buffFrames[plate][i].spellName and P.spellOpts[buffFrames[plate][i].spellName] and P.spellOpts[buffFrames[plate][i].spellName].cooldownSize or P.cooldownSize)
		end
	end
end



----------------------------------
function core:ResetStackSizes()	--
-- Update stack text size.		--
----------------------------------
	for plate in pairs(buffFrames) do 
		for i=1, table_getn(buffFrames[plate]) do 
			SetStackSize(buffFrames[plate][i], buffFrames[plate][i].spellName and P.spellOpts[buffFrames[plate][i].spellName] and P.spellOpts[buffFrames[plate][i].spellName].stackSize or P.stackSize)
--~ 			buffFrames[plate][i].stack:SetFont("Fonts\\FRIZQT__.TTF", P.stackSize, "OUTLINE")
		end
	end
end

--------------------------------------
function core:ResetAllBarPoints()	--
-- Reset all bar anchors.			--
--------------------------------------
	local barPoint		= P.barAnchorPoint
	local parentPoint	= P.plateAnchorPoint

	if P.barGrowth == 1 then --up
		barPoint = string_gsub(barPoint, "TOP", "BOTTOM");
		parentPoint = string_gsub(parentPoint, "BOTTOM", "TOP");
	else
		barPoint = string_gsub(barPoint, "BOTTOM,", "TOP");
		parentPoint = string_gsub(parentPoint, "TOP", "BOTTOM");
	end

	for plate in pairs(buffBars) do 
		if buffBars[plate][1] then
			self:ResetBarPoint(buffBars[plate][1], plate)
		end
		
		for r=2, table_getn(buffBars[plate]) do 
			buffBars[plate][r]:ClearAllPoints()
			buffBars[plate][r]:SetPoint(barPoint, buffBars[plate][r-1], parentPoint, 0, 0)
		end
	end
end

------------------------------------------------------------------
function core:ShowAllKnownSpells()								--
-- When we change number of icons to show we hide all icons.	--
-- This will reshow the buffs again in their new locations.		--
------------------------------------------------------------------
	local GUID
	for plate in pairs(buffFrames) do 
		local GUID = LibNameplate:GetGUID(plate)
		if GUID then
			self:AddBuffsToPlate(plate, GUID)
		else
			local plateName =  LibNameplate:GetName(plate)
			if plateName and nametoGUIDs[plateName] then
				self:AddBuffsToPlate(plate, nametoGUIDs[plateName])
			end
		end
	end
end